(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory SubList
imports List2 Set Multiset2

begin

inductive sublist :: "'a list \<Rightarrow> 'a list \<Rightarrow> bool" where
  Equal: "sublist [] []"
| Skip : "\<And>as b bs. sublist as bs \<Longrightarrow> sublist as (b # bs)"
| Shift: "\<And>a as bs. sublist as bs \<Longrightarrow> sublist (a # as) (a # bs)"

lemma sublist_bottom[simp]: "sublist [] as"
proof (induct as)
case Nil thus ?case by (rule Equal)
next
case (Cons a as) thus ?case by (rule Skip)
qed

lemma sublist_bottom_unique:
  assumes "sublist as []"
  shows "as = []"
using assms by (cases rule: sublist.cases) auto

lemma length_sublist:
  assumes "sublist as bs"
  shows "length as \<le> length bs"
using assms by (induct rule: sublist.induct) auto

lemma sublist_refl[simp]: "sublist as as"
proof (induct as)
case Nil thus ?case by (rule Equal)
next
case (Cons a as) thus ?case by (rule Shift)
qed

lemma prefix_sublist: "sublist as (as @ bs)"
proof (induct as)
case Nil thus ?case by simp
next
case (Cons a as) thus ?case by (simp, rule Shift)
qed

lemma suffix_sublist: "sublist bs (as @ bs)"
proof (induct as)
case Nil thus ?case by simp
next
case (Cons a as) thus ?case by (simp, rule Skip)
qed

lemma remove1_sublist: "sublist (remove1 x as) as"
proof (induct as)
case Nil thus ?case by simp
next
case (Cons a as)
  show ?case 
  proof (cases "x = a") 
  case True with Skip[of as as] show ?thesis by simp
  next
  case False with Shift[OF Cons] show ?thesis by simp
  qed
qed

lemma filter_sublist: "sublist [y\<leftarrow> as. f y] as"
proof (induct as)
case Nil thus ?case by simp
next
case (Cons a as)
  show ?case 
  proof (cases "f a") 
  case True with Shift[OF Cons] show ?thesis by simp
  next
  case False with Skip[OF Cons] show ?thesis by simp
  qed
qed

lemma map_sublist_f_g:
  assumes sublst: "sublist xs ys"
  and f_eq_g: "\<forall> x \<in> set xs. f x = g x"
  shows "sublist (map f xs) (map g ys)"
using assms proof (induct ys, (auto iff: Skip Shift)) qed
  
lemma map_sublist:
  assumes sublst: "sublist xs ys" shows "sublist (map f xs) (map f ys)"
  using assms proof (rule map_sublist_f_g[of xs ys f f], simp)  qed

lemma sublist_trans:
  assumes "sublist as bs"
  and "sublist bs cs"
  shows "sublist as cs"
using assms(2) assms(1) proof (induct arbitrary: as rule: sublist.induct)
case (Equal as) thus ?case by assumption
next
case (Skip bs c cs as) from Skip(2)[OF Skip(3)] show ?case by (rule sublist.Skip)
next
case (Shift b bs cs as) note IH = this
  show ?case using IH(3)
  proof (cases rule: sublist.cases)
  case Skip from IH(2)[OF Skip] show ?thesis by (rule sublist.Skip)
  next
  case (Shift as) from IH(2)[OF Shift(2)] Shift(1) show ?thesis by (simp add: sublist.Shift)
  qed
qed

lemma sublist_append:
  assumes "sublist as cs"
  and "sublist bs ds"
  shows "sublist (as @ bs) (cs @ ds)"
using assms(1) proof (induct rule: sublist.induct)
case Equal from assms(2) show ?case by simp
next
case (Skip as c cs)
  from sublist.Skip[OF Skip(2)] show ?case by simp
next
case (Shift a as cs)
  from sublist.Shift[OF Shift(2)] show ?case by simp
qed

lemma sublist_subset:
  assumes "sublist as bs"
  shows "set as \<le> set bs"
using assms proof (induct rule: sublist.induct)
case Equal show ?case by simp
next
case (Skip as b bs) from subset_trans[OF Skip(2)] show ?case by (cases "b \<in> set bs") auto
next
case (Shift a as bs) thus ?case by (cases "a \<in> set as") auto
qed

lemma sublist_submultiset:
  assumes "sublist as bs"
  shows "multiset_of as \<le> multiset_of bs"
using assms proof (induct rule: sublist.induct)
case Equal show ?case by simp
next
case (Skip as b bs) from order_trans[OF Skip(2)] show ?case by simp
next
case (Shift a as bs) thus ?case by simp
qed

lemma left_extension_not_sublist:
  assumes "as \<noteq> []"
  shows "\<not> sublist (as @ bs) bs"
proof-
{
  assume Hf: "sublist (as @ bs) bs"
  with length_sublist[OF this] length_append assms
   have False by simp
}
thus ?thesis by fast
qed

lemma right_extension_not_sublist:
  assumes "as \<noteq> []"
  shows "\<not> sublist (bs @ as) bs"
proof-
{
  assume Hf: "sublist (bs @ as) bs"
  with length_sublist[OF this] length_append assms
   have False by simp
}
thus ?thesis by fast
qed

lemma sublist_antisym:
  assumes "sublist as bs"
  and "sublist bs as"
  shows "as = bs"
using assms proof (induct rule: sublist.induct)
case Equal thus ?case by simp
next
case (Skip as b bs) note IH = this
  show ?case using IH(3)
  proof (cases rule: sublist.cases)
  case (Skip a as)
    from sublist_trans[OF IH(1) sublist_trans[OF SubList.sublist.Skip[OF sublist_refl[of bs], of b]
     Skip(2)]] Skip(1) left_extension_not_sublist[of "[ a ]" as] show ?thesis by simp
  next
  case (Shift as)
    from sublist_trans[OF IH(1) Shift(2)] Shift(1) left_extension_not_sublist[of "[ b ]" as]
     show ?thesis by simp
  qed
next
case (Shift a as bs) note IH = this
  show ?case using IH(3)
  proof (cases rule: sublist.cases)
  case Skip
    from sublist_trans[OF Skip IH(1)] left_extension_not_sublist[of "[ a ]" bs]
     show ?thesis by simp
  next
  case Shift
    from IH(2)[OF Shift] show ?thesis by simp
  qed
qed

lemma sublist_Cons_Cons: 
  assumes "sublist (a # as) (b # bs)"
  shows "sublist as bs" 
  using assms(1) proof (cases rule: sublist.cases)
  case Skip then show ?thesis by (metis sublist.Skip sublist_refl sublist_trans)
qed (auto)

lemma sublist_induct2:
  "\<lbrakk> sublist xs ys; 
  \<And> bs. P [] bs; 
  \<And> a as bs. \<lbrakk> sublist as bs; P as bs \<rbrakk> \<Longrightarrow> P (a # as) (a # bs); 
  \<And> a as b bs. \<lbrakk> a \<noteq> b; sublist as bs; sublist (a # as) bs; P as bs; P (a # as) bs \<rbrakk> \<Longrightarrow> P (a # as) (b # bs) \<rbrakk> 
  \<Longrightarrow> P xs ys" 
proof (induct ys arbitrary: xs)
  case Nil show ?case unfolding sublist_bottom_unique[OF Nil(1)] using Nil(2) by auto
next
  case (Cons y ys')
  note Cons_ys = Cons
  note sl = Cons(2)
  note step_eq = Cons(4)
  note step_neq = Cons(5)
  show ?case proof (cases xs) 
    case Nil show ?thesis unfolding Nil using Cons.prems(2) by auto
  next
    case (Cons x xs') 
    have sl': "sublist xs' ys'" by (metis Cons sl sublist_Cons_Cons)
    from sl' have P': "P xs' ys'" using Cons_ys by auto
    show ?thesis proof (cases "x = y")
      case False 
      have sl'': "sublist (x # xs') ys'" using sl unfolding Cons using False proof (cases rule: sublist.cases, auto) qed
      then have P'': "P (x # xs') ys'" by (metis Cons.hyps Cons_ys(3) step_eq step_neq)
      show ?thesis using step_neq[OF False sl' sl'' P' P''] unfolding Cons by auto
    next
      case True 
      show ?thesis using step_eq[OF sl' P'] unfolding Cons True[symmetric] by auto
    qed
  qed
qed

end